# bashup - shell backup

#    Copyright (C) 2008  Anselm Helbig, Caspar Clemens Mierau

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# {{{ Basic concepts

# 
# CONFIGURATION BLOCKS
#
#   Configuration options are organized into blocks. For each block a
#   shell function with the name <name-of-block>_config has to be
#   defined that outputs the names of mandatory configuration variables
#   on the first and optional variables on the second line.
#
#   Configuration blocks should belong to a section. You make them aware
#   to the system with `add_config'. This is only relevant for
#   `show_config' which can then output a nicely formatted overview of
#   the available configuration options and current settings.
# 
#   A function requiring configuration should call `check_config' with
#   the name of the configuration blocks used.
#
# MODULES
# 
#   Modules are added with `add_<module-type> <module-name>'. With this
#   you promise that the entry function with the name
#   `<module-type>_<module-name>' is available and behaves according to
#   its type. `check_modules' can then be used to assert that the
#   configured modules are available.
# 
#   Source, filter and destination module function take no parameters. A
#   source module function will output to stdout but expect no input; a
#   filter module function takes input from stdin and outputs to stdout;
#   a destination module reads from stdin but does not write to stdout.
#
#   Reporting module functions take the type of event as first argument
#   (ERROR or WARNING), the remaining arguments form the message. 
#
#   Access method modules are used by the rotation logic. They must
#   provide the following functions:
#     <module-name>_list path-to-list
#     <module-name>_move pathname-from pathname-to
#     <module-name>_remove file-to-remove
#
# DEPENDENCIES
#
#   This script was tested with bash and dash, a minimal bourne shell
#   and so should be free of "bashisms". External utilities used:
#
#     PROGRAM    VERSION TESTED
#     -----------------------------------------------------------
#     openssh    4.3p2
#     curl       7.15.5
#     openssl    0.9.7e, 0.9.8g
#     rm, ls, mv, cat, test, seq, printf, basename, dirname, expr,
#       date, sort, tty
#                GNU coreutils 5.97
#     mktemp     1.5
#     
#     tar        GNU tar 1.16
#     rsync
#
#     exp        
#     grep       GNU grep 2.5.1
#     sed        GNU sed 4.1.5
#     awk        GNU awk 3.1.5
#
#     pg_dump    
#     pg_dumpall 
#     mysqldump  
#     svnadmin   
#     
#

# }}}

# {{{ CONFIGURATION DEFAULTS

# base directory of installation, for location modules
: ${INST_DIR:=$(dirname $0)}
# the hostname appears in error messages and is used by nsca
# maybe you want to use something more descriptive
: ${HOST:=$(hostname)}
# temporary directory; used for stderr-files and the oracle fifo
: ${TMPDIR:=/tmp}

# variables for the different module types
REPORTERS=""
SOURCES=""
FILTERS=""
DESTINATIONS=""
ACCESS_METHODS=""

# use the console reporter by default if connected to a terminal
: ${USE_REPORTERS:=$(tty -s 2>/dev/null && echo console)}

core_core_config() {
    echo INST_DIR HOST SOURCE
    echo LOCK_FILE
}

# }}}
# {{{ HELPER FUNCTIONS

# {{{ configuration related

# source configuration file, fail if it cannot be read
# @param configuration file name
load_config() {
    if [ ! -r "$1" ]; then
        fail "Configuration file not readable"
    fi
    . $1
}

# make configuration block known (for show_config)
# @param 1 section
# @param 2 configuration_block
add_config() {
    local section=$1
    shift
    ALL_CONFIG="${ALL_CONFIG} ${section}_$1"
    case $section in 
        reporting) 
            REPORTING_CONFIG="${REPORTING_CONFIG} ${section}_$1";;
        source)
            SOURCE_CONFIG="${SOURCE_CONFIG} ${section}_$1";;
        destination)
            DESTINATION_CONFIG="${DESTINATION_CONFIG} ${section}_$1";;
        filter)
            FILTER_CONFIG="${FILTER_CONFIG} ${section}_$1";;
        keep)
            KEEP_CONFIG="${KEEP_CONFIG} ${section}_$1";;
        core)
            CORE_CONFIG="${CORE_CONFIG} ${section}_$1";;
    esac
}
add_config core core

# show available configuration options and their values by section
# @output configuration overview
show_config() {
    echo "Available configuration options:"
    echo
    for section in CORE REPORTING SOURCE FILTER DESTINATION KEEP; do
        echo_u "for $section"
        for c in $(get_var ${section}_CONFIG); do
            show_config_1 $c
        done
        echo
    done
}

# show settings of options in a configuration block
# @param configuration block name
# @output variable names and values
show_config_1() {
    echo "Configuration for module $1:"
    ${1}_config | {
        read required
        read optional
        for v in $required; do
            echo "	*$v=$(get_var $v)"
        done
        for v in $optional; do
            echo "	 $v=$(get_var $v)"
        done
    }
}

# check for required options, fail if one required option is unset or missing
# @param configuration block name
check_config() {
    local var vars val
    for i in $*; do
        for var in $(get_config_required $i); do
            val=$(get_var $var)
            [ "X$val" = "X" ] && fail "$var unset or empty!" 
        done
    done
}    

# get required options from configuration block
# @param configuration block name
# @output required options
get_config_required() {
    ${1}_config | {
        read required
        read optional
        echo $required
    }
}

# }}}
# {{{ modules

# tries to source modules, fails if a file cannot be read
# @params list of modules
require() {
    local module_path
    for m in $*; do
        module_path=${INST_DIR}/modules/$m
        [ -r $module_path ] && . $module_path || fail "require failed."
    done
}

# add a reporter module, check for duplicates
# @param reporter module name
add_reporter() {
    REPORTERS="$(remove_elt $1 ${REPORTERS}) $1"
}

# add a source module, check for duplicates
# @param source module name
add_source() {
    SOURCES="$(remove_elt $1 ${SOURCES}) $1"
}

# add a filter module, check for duplicates
# @param filter module name
add_filter() {
    FILTERS="$(remove_elt $1 ${FILTERS}) $1"
}

# add a destination module, check for duplicates
# @param destination module name
add_destination() {
    DESTINATIONS="$(remove_elt $1 ${DESTINATIONS}) $1"
}

# add an access method module, check for duplicates
# @param access method module name
add_access_method() {
    ACCESS_METHODS="$(remove_elt $1 ${ACCESS_METHODS}) $1"
}

# asserts the selected source, filter, destination and reporter modules are available, 
# fails if not. 
check_modules() {
    local i module
    for i in SOURCE FILTER DESTINATION; do
        module=$(get_var $i)
        member $module $(get_var ${i}S) || fail "$i Module \"$module\" is not available "
    done
    for r in ${USE_REPORTERS}; do
        member $r ${REPORTERS} || fail "Configured reporter \"$r\" not available"
    done
}

# }}}
# {{{ generic helpers

# checks if a word is in a list of words
# @param needle
# @params haystack
# @returns 0 if found, 1 if not
member() {
    local i needle=$1
    shift
    for i in $*; do
        if [ "X$i" = "X$needle" ]; then
            return 0
        fi
    done
    return 1
}

# get the content of a variable
# @param variable name
# @output variable content
get_var() {
    eval "echo \$$1"
}

# remove an element from a list of words
# @param element to remove
# @params list of words
# @output list of words with the element removed
remove_elt() {
    local remove=$1
    shift
    for i in $*; do
        [ $i != ${remove} ] && echo -n "$i "
    done
    echo
}

# removes elements in list2 from list1
# list1 and list2 are read from stdin on separate lines
# @output list1 - list2
set_subtraction() {
    local superset subset
    read superset
    read subset
    for i in $subset; do
        superset=$(remove_elt $i $superset)
    done
    echo $superset
}

# determine if list2 is a subset of list1
# list1 and list2 are read from stdin on separate lines 
# @return 0 if list2 is a subset of list1, 1 otherwise
is_subset() {
    local superset subset
    read superset
    read subset
    for i in $subset; do
        member $i $superset || return 1
    done
    return 0
}

# output words that match a given pattern (shell globbing)
# you should be careful, the pattern might get expanded if
# filename generation is active (set -f)
# @param pattern to match
# @params list of words
# @output list of words that match the pattern
match() {
    local pattern=$1
    shift
    for word in $*; do
        case $word in
            $pattern) echo $word;;
        esac
    done
}

# cuts off the extension of a filename. 
# `basename' only does this only for known extensions. 
# @param filename
# @output filename with extension removed
file_basename() {
    echo $1 | sed 's/\.[^.]\+$//g'
}

# create a temporary file for logging of standard error
# @output name of the temporary file
mkstderr() {
    local stderr=$(mktemp "${TMPDIR}/backup_stderr.XXXXXX")
    [ $? -ne 0 ] && fail "Could not create temporary file for stderr"
    echo $stderr
}

# clean up temporary files used for logging standard error
stderr_cleanup_trap() {
    # turn on file name generation again
    set +f 
    rm -f ${TMPDIR}/backup_stderr.*
}
EXIT_TRAPS="$EXIT_TRAPS stderr_cleanup_trap"

# inserts date and time data according to `date'
# add other rules for filename templates here
# @param string to be expanded
# @output string with date and time data
strftime() {
    date +"$1"
}

REPORTING_ERROR=$(mktemp "${TMPDIR}/backup_reporting_error.XXXXXX")
REPORTING_WARNING=$(mktemp "${TMPDIR}/backup_reporting_warning.XXXXXX")
exit_traps() {
    for t in $EXIT_TRAPS; do
        $t
    done
    rm -f ${REPORTING_ERROR} ${REPORTING_WARNING}
}
trap exit_traps 0

# }}}
# {{{ text output

# outputs a string with an underline
# @params words to echo
# @output words and a nice underline
echo_u() {
    echo $*
    echo $* | sed 's/./-/g'
}

# show usage
# @output usage information
usage() {
    version
    echo "Usage: $0 [OPTIONS] FILE"
    echo "Backs up one service according to configuration in FILE"
    echo 
    echo "-h show this help"
    echo "-s show configuration"
    echo "-c check for missing configuration items"
    echo "-t log errors and warnings only to stderr"
    echo "-v print version information"
    echo
    echo "This program comes with ABSOLUTELY NO WARRANTY."
    echo "This is free software, and you are welcome to redistribute it under"
    echo " the GNU General Public License 3. See file LICENSE for details"
}

version() {
    echo "Bashup Rev. 0.3-1, last changed Mi 9. Feb 16:35:56 CET 2011."
}

# }}}

# }}}
# {{{ ERROR HANDLING

# {{{ warn, fail, console reporter

# issues a warning 
# @params warning message
warn() {
    echo 1 > $REPORTING_WARNING
    for r in ${USE_REPORTERS}; do
        reporting_${r} WARNING $*
    done
}

# issues an error and aborts
# be careful: this will not exit from the script if called from a
# subshell or within a pipe; only the subprocess will be terminated. 
# @params error message
fail() {
    echo 1 > $REPORTING_ERROR
    for r in ${USE_REPORTERS}; do
        reporting_${r} ERROR $*
    done
    exit 1
}

# issues an error
fatal() {
    echo 1 > $REPORTING_ERROR
    for r in ${USE_REPORTERS}; do
        reporting_${r} ERROR $*
    done
}    

# issues a warning but skip the given reporting mechanism
# to be used when a reporting mechanism fails
# @param reporting mechanism to exclude
# @params warning message
warn_others() {
    local reporter=$1
    local save_reporters=$USE_REPORTERS
    shift
    USE_REPORTERS=$(remove_elt $reporter $USE_REPORTERS)
    warn $*
    USE_REPORTERS=$save_reporters
}

# formats a warning or error message
# @param type of message (WARNING or ERROR)
# @params message
message_format() {
    local type=$1
    shift
    # echo "Call stack: " ${FUNCNAME[@]}
    echo "[$HOST] [$(date)] ${type}: $*"
}

# the console reporting module
# output gets sent to stderr
add_reporter console
reporting_console() {
    echo $(message_format $*) >&2
}

# }}}
# {{{ logfile writing 

# the log reporting module
add_reporter log
add_config reporting log
reporting_log_config() {
    echo COMBINED_LOG 
    echo ERROR_LOG WARN_LOG
}
reporting_log() {
    local combined=$(strftime "$COMBINED_LOG")
    local error=$(strftime "$ERROR_LOG")
    local warn=$(strftime "$WARN_LOG")

    if check_log "${combined}" combined; then
        message_format $* >> ${combined}
    fi
    if [ $1 = "ERROR" ]; then
        if check_log "${error}" error; then
            message_format $* >> ${error}
        fi
    else
        if check_log "${warn}" warn; then
            message_format $* >> ${warn}
        fi
    fi
}

reporting_log_trap() {
    if member log $USE_REPORTERS; then 
        [ "X$(cat $REPORTING_ERROR)" = "X" ] && rm -f $ERROR_LOG
        [ "X$(cat $REPORTING_WARNING)" = "X" ] && rm -f $WARN_LOG
    fi
}
EXIT_TRAPS="$EXIT_TRAPS reporting_log_trap"

# check if a logfile is writeable, issue a warning if not
# @param path to logfile
# @param type of log (optional)
check_log() {
    local log=$1
    local type=$2
    if [ -w $log ] || touch "$log"; then
        return 0
    else
        warn_others log "Could not write to ${type} log \"${log}\""
        return 1
    fi
}

# }}}
# {{{ nagios reporting with nsca

# nsca reporting module
# FIXME: untested
add_reporter nsca
add_config reporting nsca
reporting_nsca_config() {
    echo NSCA_SERVICE NSCA_SERVER NSCA_CONFIG_FILE
    echo NSCA_HOST SEND_NSCA_PROGRAM NSCA_TIMEOUT
}
reporting_nsca() {
    local type=$1
    local return_code=2
    [ $type == "WARNING" ] && return_code=1
    shift
    reporting_nsca_send $return_code $*
}

# report success when no warnings have been sent
# to be used in an exit trap
reporting_nsca_trap() {
    if member nsca $USE_REPORTERS; then
        if [ "X$(cat $REPORTING_ERROR)" = "X" ] && [ "X$(cat $REPORTING_WARNING)" = "X" ]; then
            reporting_nsca_send 0 "Backup succeeded without errors or warnings."
        fi
    fi
}
EXIT_TRAPS="$EXIT_TRAPS reporting_nsca_trap"

reporting_nsca_send() {
    local return_code=$1
    shift
    stderr=$(mkstderr)
    # the string to echo contains literal TABs, as there seems to be no portable way to present them with "\t"
    echo "${NSCA_HOST:-${HOST}}	${NSCA_SERVICE}	${return_code}	$(message_format $*)" | \
        ${SEND_NSCA_PROGRAM:-send_nsca} -H $NSCA_SERVER ${NSCA_TIMEOUT:+-to $NSCA_TIMEOUT} -c ${NSCA_CONFIG_FILE} 2> $stderr\
        || recursive_warn nsca "Sending Nagios notification failed: $(cat $stderr)"
}

# }}}
# {{{ xmpp (jabber) reporting

# the xmpp (jabber) reporting module
# /usr/share/perl5/XML/Stream.pm must be patched for sendxmpp to work with the openfire jabber server
# around line 1163, change
#   delete($self->{SIDS}->{$currsid});
# to 
#   delete($self->{SIDS}->{$currsid}) unless ($currsid eq $sid);
add_reporter xmpp
add_config reporting xmpp
reporting_xmpp_config() {
    echo XMPP_RECIPIENTS
    echo XMPP_USER XMPP_PASSWORD XMPP_SERVER XMPP_RESOURCE SENDXMPP_PROGRAM
}
reporting_xmpp() {
    stderr=$(mkstderr)
    message_format $* | \
        ${SENDXMPP_PROGRAM:-sendxmpp} ${XMPP_USER:+-u $XMPP_USER} \
        ${XMPP_PASSWORD:+-p $XMPP_PASSWORD} ${XMPP_SERVER:+-j $XMPP_SERVER} \
        ${XMPP_RESOURCE:+-r $XMPP_RESOURCE} -t ${XMPP_RECIPIENTS} 2> $stderr \
        || recursive_warn xmpp "Sending XMPP message failed: $(cat $stderr)"
}

# }}}
# {{{ mail reporting

# the mail reporting module
# FIXME: untested
add_reporter mail
add_config reporting mail
reporting_mail_config() {
    echo MAIL_RECIPIENTS
    echo MAIL_PROGRAM
}

reporting_mail() {
    stderr=$(mkstderr)
    message_format $* | ${MAIL_PROGRAM:-mail} -s "$(message_format $*)" ${MAIL_RECIPIENTS} 2> $stderr \
        || recursive_warn mail "Sending mail failed: $(cat $stderr)"
}   

# }}}

# }}}
# {{{ DATA SOURCES

# {{{ single file

add_source file
add_config source file
source_file_config() {
    echo SOURCE_FILE
}
source_file() {
    cat ${SOURCE_FILE}
}

# }}}
# {{{ directory tree

# the tree source module
# outputs an uncompressed tar archive
add_source tree
add_config source tree
source_tree_config() {
    echo TREE_DIRECTORIES
    echo TREE_EXCLUDES
}
source_tree() {
    local exclude_list=""
    local directory_list=""

    check_config source_tree
    if [ ! "X${TREE_EXCLUDES}" = "X" ]; then
        for e in ${TREE_EXCLUDES}; do
            exclude_list="$exclude_list --exclude=$e"
        done
    fi
    # make absolute paths relative, lest tar will complain
    for d in $TREE_DIRECTORIES; do
        directory_list="$directory_list $(echo $d | sed 's|^/||')"
    done

    stderr=$(mkstderr)
    tar -cf - -C / $exclude_list $directory_list 2> $stderr
    case $? in
        1) warn "tar had warnings: $(cat $stderr)";;
        2) fail "tar failed: $(cat $stderr)";;
    esac
}

# }}}
# {{{ mysql 

add_config source mysql
source_mysql_config() {
    echo MYSQL_USER
    echo MYSQL_PASSWORD MYSQL_HOST MYSQL_DATABASES MYSQL_PROGRAM MYSQL_TABLE_EXCLUDE_PATTERN
}

source_mysqldump_config() {
    echo
    echo MYSQLDUMP_PROGRAM MYSQLDUMP_OPTIONS 
}

# mysqldump source module
add_source mysqldump
source_mysqldump() {
    local options=--all-databases
    local db query
    check_config source_mysql
    check_config source_mysqldump
    if [ "X${MYSQL_DATABASES}" != "X" ]; then
        if [ "X$MYSQL_TABLE_EXCLUDE_PATTERN" != "X" ]; then
            db=${MYSQL_DATABASES%% *}
            query="SELECT table_name FROM INFORMATION_SCHEMA.TABLES WHERE table_schema = '${db}' AND table_name NOT LIKE '${MYSQL_TABLE_EXCLUDE_PATTERN}'"
            options="$db $(${MYSQL_PROGRAM:-mysql} -u ${MYSQL_USER} ${MYSQL_PASSWORD:+-p${MYSQL_PASSWORD}} ${MYSQL_HOST:+-h ${MYSQL_HOST}} --batch --skip-column-names --execute="$query")"
        else
            options="--databases ${MYSQL_DATABASES}"
        fi
    fi
    stderr=$(mkstderr)
    ${MYSQLDUMP_PROGRAM:-mysqldump} -u ${MYSQL_USER} ${MYSQL_PASSWORD:+-p${MYSQL_PASSWORD}} ${MYSQL_HOST:+-h ${MYSQL_HOST}} ${MYSQLDUMP_OPTIONS} $options 2> $stderr
    [ $? -ne 0 ] && fail "mysqldump failed: $(cat $stderr)"
}

# }}}
# {{{ postgresql

add_config source postgresql
source_postgresql_config() {
    echo 
    echo PSQL_USER PSQL_PORT PSQL_DATABASES
}

# postgresql source module
# the backup user must be able to log in without a password
add_source postgresql
source_postgresql() {
    check_config source_postgresql
    stderr=$(mkstderr)
    if [ "X${PSQL_DATABASES}" != "X" ]; then
        pg_dump ${PSQL_USER:+-U $PSQL_USER} ${PSQL_PORT:+-p $PSQL_PORT} ${PSQL_DATABASES} 2> $stderr
    else
        pg_dumpall ${PSQL_USER:+-U $PSQL_USER} ${PSQL_PORT:+-p $PSQL_PORT} 2> $stderr
    fi
    [ $? -ne 0 ] && fail "pg_dump failed: $(cat $stderr)"
}

# }}}
# {{{ oracle

add_config source oracle
source_oracle_config() {
    echo ORACLE_USER ORACLE_PASSWORD ORACLE_CONNECTION
    echo ORACLE_DATABASES ORACLE_LOG ORACLE_USER_ONLY ORACLE_EXP_PROGRAM
}

# the oracle exp source module
add_source oracle
source_oracle() {
    local databases=""
    local log=""

    check_config source_oracle

    if [ "X${ORACLE_DATABASES}" = "X" ]; then
        if [ "X" = "X${ORACLE_USER_ONLY}" ]; then
            databases="full=y"
        fi
    else
        for db in ${ORACLE_DATABASES}; do
            databases="$databases${db},"
        done
        databases="owner=${databases%,}"
    fi

    if [ "X${ORACLE_LOG}" != "X" ]; then
        log="$(strftime ${ORACLE_LOG})"
        if check_log $log oracle; then
            log="log=$log"
        else
            log=""
        fi
    fi

    # output gets written to a named pipe since oracle can not export to stdout

    # create a named pipe or use an existing one
    fifo=${ORACLE_FIFO:-${TMPDIR}/oracle_fifo}
    ([ -p ${fifo} ] && [ -w ${fifo} ]) || mkfifo -m 0600 ${fifo} || fail "Could not create fifo \"$fifo\" for oracle export."

    stderr=$(mkstderr)

    # the export gets executed in the background in a subshell because
    # the output from the fifo has to run in the foreground. When the
    # export fails without trying to open the fifo, the foreground
    # process never returns. FIXME: can we do something about it?
    (${ORACLE_EXP_PROGRAM:-exp} ${ORACLE_USER}/${ORACLE_PASSWORD}@${ORACLE_CONNECTION} \
        $databases statistics=none consistent=y $log file=${fifo} 2> $stderr
        oracle_check_success $? $stderr) &

    # read from the fifo and write to stdout until the fifo is closed
    cat < ${fifo}
    # clean up
    rm -f ${fifo}
}
# extracts warning messages from oracle exp output.
# consecutive repeated messages get collapsed into one.
# @param filename
oracle_extract_warnings() {
    grep -E '^[A-Z]{3}' $1 | uniq -c | \
        awk '{ if ($1 > 1) printf "%sx ", $1; printf "\""; for (i=2; i <= NF; i++) printf "%s ", $i; print "\"; " }'
}
# check for success of oracle exp
# @param exit status
# @param file with stderr output
# @throws error when exp unsuccessful or unknown, warn if there were
#         warnings or the stderr output suggests all were ok, yet the
#         exit status is non-zero
oracle_check_success() {
    local status=$1
    local log=$2
    case $(tail -1 $log) in
        *"unsuccessfully")
            fail "Oracle exp failed with exit code $status: $(oracle_extract_warnings $log)";;
        *"successfully with warnings.")
            warn "Oracle exp had warnings, exit code $status: $(oracle_extract_warnings $log)";;
        *"successfully without warnings.")
            if [ $status -ne 0 ]; then
                warn "Oracle log had no warnings, but exited with code $status: $(oracle_extract_warnings $log)"
            fi
            ;;
        *)
            fail "Oracle exp exited with unknown message, exit code $status: $(oracle_extract_warnings $log)";;
    esac
}

# }}}
# {{{ subversion

add_config source subversion
source_subversion_config() {
    echo SUBVERSION_REPOSITORY_ROOT
}

add_config source subversion_hotcopy
source_subversion_hotcopy_config() {
    echo SUBVERSION_DESTINATION
    echo SUBVERSION_INCLUDES SUBVERSION_EXCLUDES SUBVERSION_TMPDIR
}

# subversion dump source
# FIXME: untested
add_source subversion_dump
source_subversion_dump() {
    check_config source_subversion
    stderr=$(mkstderr)
    svnadmin dump ${SUBVERSION_REPOSITORY} 2> $stderr || fail "svnadmin dump failed: $(cat $stderr)"
}

# subversion hotcopy source
# FIXME: untested
add_source subversion_hotcopy
source_subversion_hotcopy() {
    local tmpdir repositories e r

    check_config source_subversion
    check_config source_subversion_hotcopy
    dest_dir=$(strftime ${DIRECTORY})
    mkdir -p ${dest_dir}
    check_config destination_directory

    stderr=$(mkstderr)
    tmpdir=${SUBVERSION_TMPDIR:-${TMPDIR}/subversion-hotcopy}
    rm -rf ${tmpdir} 2> ${stderr} || fail "svn hotcopy: could not clean  temporary directory ${tmpdir}: $(cat $stderr)"
    mkdir -p ${tmpdir} 2> ${stderr} || fail "svn hotcopy: could not create temporary directory ${tmpdir}: $(cat $stderr)"

    repositories="${SUBVERSION_INCLUDES:-$(ls ${SUBVERSION_REPOSITORY_ROOT})}"
    for e in ${SUBVERSION_EXCLUDES}; do
        repositories=$(remove_elt $e $repositories)
    done


    for r in ${repositories}; do 
        svnadmin hotcopy ${SUBVERSION_REPOSITORY_ROOT}/${r} ${tmpdir}/${r} 2> $stderr || fail "svnadmin hotcopy failed: $(cat $stderr)"
	# as we are creating more than one file we have to decide which filter to user here, too
	case ${FILTER} in
		gzip)
		    FILE="${dest_dir}/${r}.tar.gz";;
		bzip2)
		    FILE="${dest_dir}/${r}.tar.bz2";;
		lzma)
		    FILE="${dest_dir}/${r}.tar.lzma";;
		*)
		    FILE="${dest_dir}/${r}.tar";;
	esac
        tar cf - -C ${tmpdir} ${r} | filter_${FILTER} | destination_${SUBVERSION_DESTINATION:-none}
        rm -rf ${tmpdir}/${r}
    done
    rmdir ${tmpdir}
}

# }}}

# }}}
# {{{ BACKUP DESTINATIONS

# {{{ none

# /dev/null destination module; space efficient and fast.
# seriously: this is only useful for testing and for sources like
# rsync that save the data themselves
add_destination none
destination_none() {
    cat > /dev/null
}

# }}}
# {{{ file, directory

add_config destination file
destination_file_config() {
    echo FILE
}

destination_directory_config() {
    echo DIRECTORY
}

# (local) file destination module
add_destination file
destination_file() {
    local f

    check_config destination_file
    f="$(strftime ${FILE})"
    [ -w ${f} ] || touch ${f} || fail "Destination file ${f} not writable"
    cat > ${f}
}

# }}}
# {{{ ssh

add_config destination ssh
destination_ssh_config() {
    echo SSH_HOST
    echo SSH_USER SSH_PROGRAM
}

# ssh_file destination module
add_destination ssh_file
destination_ssh_file() {
    check_config destination_file destination_ssh
    stderr=$(mkstderr)
    ${SSH_PROGRAM:-ssh} ${SSH_USER:+-l $SSH_USER} ${SSH_HOST} "cat > $(strftime ${FILE})" 2> $stderr \
        || fail "ssh failed: $(cat $stderr)"
}

# ssh_directory destination module
# expects an uncompressed tar on stdin
add_destination ssh_directory
destination_ssh_directory() {
    check_config destination_directory destination_ssh
    stderr=$(mkstderr)
    ${SSH_PROGRAM:-ssh} ${SSH_USER:+-l $SSH_USER} ${SSH_HOST} "tar xf - -C $(strftime ${DESTINATION_DIRECTORY})" 2> $stderr \
        || fail "ssh failed: $(cat $stderr)"
}

# }}}
# {{{ ftp

add_config destination ftp
destination_ftp_config() {
    echo FTP_HOST
    echo CURL_PROGRAM CURL_OPTIONS
}

# ftp_file destination module
add_destination ftp_file
destination_ftp_file() {
    check_config destination_file destination_ftp
    stderr=$(mkstderr)
    
    f="$(strftime ${FILE})"
    ${CURL_PROGRAM:-curl} ${CURL_OPTIONS} \
        --netrc-optional --silent --show-error --upload-file - --ftp-create-dirs \
        ftp://${FTP_HOST}/${f} 2> $stderr \
        || fail "curl failed: $(cat $stderr)"
}

# }}}
# {{{ rsync

add_config destination rsync
destination_rsync_config() {
    echo
    echo RSYNC_PROGRAM RSYNC_OPTIONS
}

# rsync destination module
# FIXME: untested
# FIXME: does not make much sense as a module, we should rather
#        include the error reporting as a library from a user-written
#        script
add_destination rsync_file
destination_rsync_file() {
    check_config destination_file destination_rsync
    stderr=$(mkstderr)
    ${RSYNC_PROGRAM:-rsync} ${RSYNC_OPTIONS} 2> $stderr \
        || fail "rsync failed: $(cat $stderr)"
}

# }}}
# {{{ s3

add_config destination s3
destination_s3_config() {
    echo S3_ACCESS_KEY_ID S3_SECRET_ACCESS_KEY S3_BUCKET
    echo S3_URL OPENSSL_PROGRAM BASE64_PROGRAM S3_CONTENT_TYPE CURL_PROGRAM
}

add_destination s3
destination_s3() {
    check_config destination_file destination_s3
    # TODO: check if we can handle s3 uploads without a temporary file
    local temp_file="$(mktemp "${TMPDIR}/s3_upload.XXXXXX")"
    [ $? -ne 0 ] && fail "Could not create temporary file for s3 upload"
    cat > "$temp_file"
    _s3_request "PUT" "$FILE" "$temp_file"
    # TODO: check errors from AWS
    rm -f "$temp_file"
}
destination_s3_file() {
    check_config destination_file destination_s3
    _s3_request "PUT" "$FILE" "$1"
}

_s3_request() {
    local method="$1" url="/${S3_BUCKET}/$2" file="$3" amz_headers="$4"
    local content_type="${S3_CONTENT_TYPE:-application/octet-stream}"
    # FIXME: date  +"%a, %d %b %Y %H:%M:%S %z" might be more portable
    local date="$(date --rfc-2822)"
    local curl_options=""
    local signature=$(
        ( echo "$method"
          echo ""
          echo "$content_type"
          echo "$date"
          if [ "X${amz_headers}" != "X" ]; then
              echo "${amz_headers}" | sed 's/: /:/'
          fi
          echo -n "$url" ) | \
          ${OPENSSL_PROGRAM:-openssl} sha1 -hmac "${S3_SECRET_ACCESS_KEY}" -binary | ${OPENSSL_PROGRAM:-openssl} base64)
    local stderr="$(mkstderr)"
    local stdout="$(mkstderr)"
    ${CURL_PROGRAM:-curl} ${file:+--upload-file "$file"} \
        --request "$method" \
        --header "Authorization: AWS ${S3_ACCESS_KEY_ID}:$signature" \
        --header "Date: $date" \
        --header "$amz_headers" \
        --header "Content-Type: ${content_type}" \
        --header "Transfer-Encoding:" \
        "${S3_URL:-https://s3.amazonaws.com}$url" 2> $stderr > $stdout \
        || fail "curl failed: $(cat $stderr)"
    grep -q '<Error>' $stdout && fail "S3 request failed: $(cat $stdout)"
    cat $stdout
}

# }}}

# }}}
# {{{ FILTERS

add_filter gzip
filter_gzip() {
    gzip -c 
}

add_filter bzip2
filter_bzip2() {
    bzip2 -c
}

add_filter lzma
filter_lzma() {
    lzma -c
}

add_config filter gpg
filter_gpg_config() {
    echo 
    echo GPG_PROGRAM GPG_PASSPHRASE_FILE GPG_PASSPHRASE
}
add_filter gpg_symmetric
filter_gpg_symmetric() {
    check_config filter_gpg
    ${GPG_PROGRAM:-gpg} --symmetric --no-use-agent ${GPG_PASSPHRASE_FILE:+--passphrase-file "${GPG_PASSPHRASE_FILE}"} ${GPG_PASSPHRASE:+--passphrase "${GPG_PASSPHRASE}"}
    # TODO: use `gpg --decrypt --no-use-agent --passphrase...' to decrypt
}

# other possibly useful filters: base64, uuencode, encryption, ...

# }}}
# {{{ ROTATION

# disable globbing - we want to do it ourselves
set -f

rotate_config() {
    echo ROTATE_ACCESS_METHOD ROTATE_LOCATION
    echo ROTATE_KEEP ROTATE_MOVE
}

# handles deletion and renaming of backups
rotate() {
    local files files_to_keep

    if [ "X$ROTATE_ACCESS_METHOD" = "X" ]; then
        return
    fi
    if [ "X$ROTATE_LOCATION" = "X" ]; then
        fail "Your ROTATE_LOCATION is empty - please set one matching all your files you want to rotate"
    fi

    check_config rotate

    local location_dir=$(dirname $ROTATE_LOCATION)
    local location_match=$(basename $ROTATE_LOCATION)
    local location_base=$(file_basename $location_match)


    # assert that access method is registered
    member $ROTATE_ACCESS_METHOD $ACCESS_METHODS || fail "Access method \"$ROTATE_ACCESS_METHOD\" not available"

    # determine the list of files matching the pattern
    files=$(match $location_match $(${ROTATE_ACCESS_METHOD}_list $location_dir))
    files_to_keep="${files}"
    
    if [ "X${ROTATE_KEEP}" != "X" ]; then
        files_to_keep=""
        # call keep functions
        for kf in ${ROTATE_KEEP}; do 
            check_config ${kf}
            files_to_keep="$files_to_keep $($kf $files)"
        done

        # remove files we don't want to keep
        for f in $({ echo ${files}; echo ${files_to_keep}; } | set_subtraction); do
            ${ROTATE_ACCESS_METHOD}_remove ${location_dir}/$f
        done
    fi
    
    if [ "X$ROTATE_MOVE" != "X" ]; then
        ${ROTATE_MOVE} $location_dir $location_base $files_to_keep
    fi
}

# {{{ numbered backups

# handles numbered backups
# @param  base directory
# @params filenames
move_numbered() {
    local location_dir=$1
    local location_base=$2
    shift 
    shift
    # sort filenames in descending order before renaming, so we don't
    # incidentially overwrite backups
    for f in $(echo $* | sed 's/ \+/\n/g' | sort -r); do
        ${ROTATE_ACCESS_METHOD}_move ${location_dir}/$f ${location_dir}/$(next_filename $f)
    done
    # rename the backup file without a number 
    ${ROTATE_ACCESS_METHOD}_move ${location_dir}/$location_base ${location_dir}/${location_base}.001
}

# takes a filename of the form <somename>.NNN and gives the next
# filename, increasing NNN by one
# @param filename
# @outputs next filename
next_filename() {
    local fn=$1
    local base=$(file_basename $fn)
    local ext=$(echo $fn | sed -n 's/.*\.\([^\.]\+\)$/\1/p')
    local next=$(expr $ext + 1)
    if [ $? -eq 0 ]; then
        echo ${base}.$(printf %03d $next)
    else
        warn "Could not determine next filename from \"$fn\""
        echo $fn
    fi
}

# }}}
# {{{ access method modules

# {{{ debug

# debug access method module; for testing the renaming logic
add_access_method debug
debug_list() {
    # for i in $(seq 10); do printf "bu.%03d " $i; done; echo bu
    # for i in $(seq 100); do echo bu.$(date -I --date="-$i day"); done
    echo bu bu.001 bu.010 bu.011 bu.012 bu.013 bu.014 bu.015 bu.016 bu.017 bu.018 bu.019 bu.002 bu.025 bu.003 bu.004 bu.005 bu.006 bu.007 bu.008 bu.009
}
debug_move() {
    echo mv $1 $2
}
debug_remove() {
    echo rm $1
}

# }}}
# {{{ local

# local access method module
add_access_method local
local_list() {
    local stderr=$(mkstderr)
    ls $1 2> $stderr || fail "local list failed: $(cat $stderr)"
}
local_move() {
    local from=$1 to=$2
    local stderr=$(mkstderr)
    mv $1 $2 2> $stderr || fail "local move failed: $(cat $stderr)"
}
local_remove() {
    local stderr=$(mkstderr)
    rm $1 2> $stderr || fail "local remove failed: $(cat $stderr)"
}

# }}}
# {{{ ftp

# ftp access method module

add_access_method ftp
ftp_list() {
    local stderr=$(mkstderr)
    check_config destination_ftp
    ${CURL_PROGRAM:-curl} ${CURL_OPTIONS} \
        --list-only --netrc-optional \
        ftp://${FTP_HOST}/${1}/ 2> $stderr | sed "s/\r$//g" \
        || fail "curl list failed: $(cat $stderr)"
}
ftp_move() {
    local from=$1 to=$2
    local stderr=$(mkstderr)

    check_config destination_ftp
    ${CURL_PROGRAM:-curl} ${CURL_OPTIONS} \
        --netrc-optional --silent --show-error \
        --quote "RNFR $from" --quote "RNTO $to" \
        ftp://${FTP_HOST}/ > /dev/null 2> $stderr \
        || fail "curl move \"$from\" -> \"$to\" failed: $(cat $stderr)"
}
ftp_remove() {
    local stderr=$(mkstderr)

    check_config destination_ftp
    ${CURL_PROGRAM:-curl} ${CURL_OPTIONS} \
        --netrc-optional --silent --show-error \
        --quote "DELE $1" \
        ftp://${FTP_HOST}/ > /dev/null 2> $stderr \
        || fail "curl remove failed: $(cat $stderr)"
}

# }}}
# {{{ ftp_directories

add_access_method ftp_directories
ftp_directories_list() {
    ftp_list $*
}
ftp_directories_move() {
    ftp_move $*
}
ftp_directories_remove() {
    local stderr=$(mkstderr)
    check_config destination_ftp
    for f in $(ftp_list $1); do
        [ $f != "."  -a $f != ".." ] && ftp_remove ${1}/${f}
    done
    ${CURL_PROGRAM:-curl} ${CURL_OPTIONS} \
        --netrc-optional --silent --show-error \
        --quote "RMD $1" \
        ftp://${FTP_HOST}/ > /dev/null 2> $stderr \
        || fail "curl rmdir failed: $(cat $stderr)"
}

# }}}
# {{{ ssh

add_access_method ssh
ssh_list() {
    local stderr=$(mkstderr)
    check_config destination_ssh
    ${SSH_PROGRAM:-ssh} ${SSH_USER:+-l $SSH_USER} ${SSH_HOST} "ls $1" 2> $stderr \
        || fail "ssh list failed: $(cat $stderr)"
}
ssh_move() {
    local from=$1 to=$2
    local stderr=$(mkstderr)

    ${SSH_PROGRAM:-ssh} ${SSH_USER:+-l $SSH_USER} ${SSH_HOST} "mv $1 $2" 2> $stderr \
        || fail "ssh move failed: $(cat $stderr)"
}
ssh_remove() {
    local stderr=$(mkstderr)
    ${SSH_PROGRAM:-ssh} ${SSH_USER:+-l $SSH_USER} ${SSH_HOST} "rm $1" 2> $stderr \
        || fail "ssh remove failed: $(cat $stderr)"
}    

# }}}
# {{{ s3

add_access_method s3
s3_list() {
    check_config destination_s3
    _s3_request "GET" "$1" | sed 's|</Key>|\0\n|g' | sed -n 's|.*<Key>\([^>]*\)</Key>.*|\1|gp'
}
s3_move() {
    check_config destination_s3
    _s3_request "PUT" "$2" "" "x-amz-copy-source: /${S3_BUCKET}/$1"
    s3_remove "$1"
}
s3_remove() {
    check_config destination_s3
    _s3_request "DELETE" "$1"
}

# }}}

# }}}
# {{{ keep functions

# {{{ numbered backups

add_config keep numbered
keep_numbered_config() {
    echo FILES_TO_KEEP
}
keep_numbered() {
    for file in $*; do
        [ "$(get_index $file)" -le $FILES_TO_KEEP ] \
            && echo $file
    done
}
get_index() {
    echo $* | sed -n 's/.*\([0-9]\{3\}\).*/\1/p'
    return $?
}

# }}}
# {{{ backups by date

add_config keep daily
keep_daily_config() {
    echo DAYS_TO_KEEP
}
keep_daily() {
    keep_by_date keep_daily_predicate $*
}
keep_daily_predicate() {
    [ $(expr $2 * 365 + $(date +%j) - $(date --date=$1 +%j)) -lt $DAYS_TO_KEEP ]
    return $?
}

add_config keep weekly
keep_weekly_config() {
    echo WEEKDAY_TO_KEEP WEEKS_TO_KEEP
}
keep_weekly() {
    keep_by_date keep_weekly_predicate $*
}
keep_weekly_predicate () {
    [ $(date --date=$1 +%w) -eq $WEEKDAY_TO_KEEP ] && \
        [ $(expr $2 * 54 + $(date +%U) - $(date --date=$1 +%U) - "(" $WEEKDAY_TO_KEEP ">=" $(date +%w) ")" ) -lt $WEEKS_TO_KEEP ]
    return $?
}

add_config keep monthly
keep_monthly_config() {
    echo MONTHDAY_TO_KEEP MONTHS_TO_KEEP
}
keep_monthly() {
    keep_by_date keep_monthly_predicate $*
}
keep_monthly_predicate () {
    [ $(date --date=$1 +%d) -eq $MONTHDAY_TO_KEEP ] && \
        [ $(expr $2 * 12 + $(date +%m) - $(date --date=$1 +%m) - "(" $MONTHDAY_TO_KEEP ">=" $(date +%d) ")" ) -lt $MONTHS_TO_KEEP ]
    return $?
}

# tries to extract an iso date from its arguments
# @params filename
# @output extracted iso date
# @return 0 on success, sed exit code on failure
get_date() {
    echo $* | sed -n 's/.*\([0-9]\{4\}-[0-9]\{1,2\}-[0-9]\{1,2\}\).*/\1/p'
    return $?
}

keep_by_date() {
    local date file year_change
    local keep_predicate=$1
    shift
    for file in $*; do
        date=$(get_date $file)
        year_change=$(expr \( $(date +%Y) - $(date --date=$date +%Y) \) )
        $keep_predicate $date $year_change && echo $file
    done
}

# }}}

# }}}

# }}}
# {{{ MAIN

main() {
    while getopts "hvsct" arg; do
        case $arg in
            h|\?) 
                usage; exit 0;;
            v) 
                version; exit 0;;
            s) 
                shift $(expr $OPTIND - 1); [ $# -ge 1 ] && load_config $1; show_config; exit;;
            c) 
                shift $(expr $OPTIND - 1); load_config $1; check_config ${ALL_CONFIG}; exit;;
            t)
                CONSOLE_ONLY=1
        esac
    done
    
    shift $(expr $OPTIND - 1)

    if [ $# -lt 1 ]; then
        usage
        exit
    fi
    
    acquire_lock
    . $1
    bashup

}

filter() {
    local filters="$*" cmd=""
    for f in $filters; do
        cmd="$cmd filter_$f | "
    done
    eval "${cmd}cat"
}

bashup() {

    if [ "X$CONSOLE_ONLY" != "X" ]; then
        USE_REPORTERS="console"
    fi

    DESTINATION=${DESTINATION:-none}

    check_config core_core
    check_modules

    # we deliberately execute the rotate in a subshell, so that
    # failures while rotating do not prevent the actual backup from
    # happening
    (rotate)

    source_${SOURCE} | filter ${FILTER} | destination_${DESTINATION}
}



acquire_lock() {
    if [ "X$LOCK_FILE" != "X" ]; then
        [ -e $LOCK_FILE ] && fail "Another backup process is running, PID $(cat $LOCK_FILE); if the process is dead, remove \"$LOCK_FILE\""
        if echo $$ > $LOCK_FILE; then
            EXIT_TRAPS="$EXIT_TRAPS remove_lock"
        else
            warn "Could not create lock file \"$LOCK_FILE\""
        fi
    fi
}
remove_lock() {
    rm -f ${LOCK_FILE} || warn "Could not remove lock file \"$LOCK_FILE\"!"
}

# }}}

# vim: set foldmethod=marker:
# Local Variables:
# mode: folding
# End:
